#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Connect various satpy components together to go from satellite data to output imagery format."""

from __future__ import annotations

import contextlib
import logging
import os
import shutil
import sys
import tempfile
from collections.abc import Iterable
from datetime import datetime
from typing import Optional, Union

# isort: off
# hdf5plugin must be imported before h5py and xarray or it won't be available
# Used by the FCI reader
try:
    import hdf5plugin
except ImportError:
    hdf5plugin = None  # type: ignore
# isort: on

import dask
import satpy
from dask.diagnostics import ProgressBar
from pyresample import SwathDefinition
from satpy import DataID, Scene
from satpy.writers import compute_writer_results

from polar2grid._glue_argparser import GlueArgumentParser, get_p2g_defaults_env_var
from polar2grid.core.script_utils import create_exc_handler, rename_log_file, setup_logging
from polar2grid.filters import filter_scene
from polar2grid.readers._base import ReaderProxyBase
from polar2grid.resample import resample_scene
from polar2grid.utils.config import add_polar2grid_config_paths
from polar2grid.utils.dynamic_imports import get_reader_attr
from polar2grid.utils.legacy_compat import get_sensor_alias

LOG = logging.getLogger(__name__)

_PLATFORM_ALIASES = {
    "suominpp": "npp",
    "npp": "npp",
    "snpp": "npp",
    "n20": "noaa20",
    "n21": "noaa21",
    "n22": "noaa22",
    "n23": "noaa23",
    "noaa18": "noaa18",
    "noaa19": "noaa19",
    "noaa20": "noaa20",
    "noaa21": "noaa21",
    "noaa22": "noaa22",
    "noaa23": "noaa23",
    "jpss1": "noaa20",
    "jpss2": "noaa21",
    "jpss3": "noaa22",
    "jpss4": "noaa23",
    "j1": "noaa20",
    "j2": "noaa21",
    "j3": "noaa22",
    "j4": "noaa23",
    "fy3b": "fy3b",
    "fy3c": "fy3c",
    "fy3d": "fy3d",
    "eosaqua": "aqua",
    "eosterra": "terra",
    "aqua": "aqua",
    "terra": "terra",
    "gcomw1": "gcom-w1",
    "metopa": "metopa",
    "metopb": "metopb",
    "metopc": "metopc",
}


def _get_platform_name_alias(satpy_platform_name):
    return _PLATFORM_ALIASES.get(satpy_platform_name.lower().replace("-", ""), satpy_platform_name)


def _overwrite_platform_name_with_aliases(scn):
    """Change 'platform_name' for every DataArray to Polar2Grid expectations."""
    for data_arr in scn:
        if "platform_name" not in data_arr.attrs:
            continue
        pname = _get_platform_name_alias(data_arr.attrs["platform_name"])
        data_arr.attrs["platform_name"] = pname


def _overwrite_sensor_with_aliases(scn):
    """Change 'sensor' for every DataArray to Polar2Grid expectations."""
    for data_arr in scn:
        if "sensor" not in data_arr.attrs:
            continue
        pname = get_sensor_alias(data_arr.attrs["sensor"])
        data_arr.attrs["sensor"] = pname


def _write_scene(
    scn: Scene, writers: list[str], writer_args: dict[str, dict], data_ids: list[DataID], to_save: Optional[list] = None
):
    if to_save is None:
        to_save = []
    if not data_ids:
        # no datasets to save
        return to_save

    _assign_default_native_area_id(scn, data_ids)
    for writer_name in writers:
        wargs = writer_args[writer_name]
        _write_scene_with_writer(scn, writer_name, data_ids, wargs, to_save)
    return to_save


def _assign_default_native_area_id(scn: Scene, data_ids: list[DataID]) -> None:
    for data_id in data_ids:
        area_def = scn[data_id].attrs.get("area")
        if area_def is None or hasattr(area_def, "area_id"):
            continue
        scn[data_id].attrs["area"].area_id = "native"


def _write_scene_with_writer(scn: Scene, writer_name: str, data_ids: list[DataID], wargs: dict, to_save: list) -> None:
    res = scn.save_datasets(writer=writer_name, compute=False, datasets=data_ids, **wargs)
    if res and isinstance(res[0], (tuple, list)):
        # list of (dask-array, file-obj) tuples
        to_save.extend(zip(*res, strict=True))
    else:
        # list of delayed objects
        to_save.extend(res)


def _print_list_products(reader_info, is_polar2grid: bool, p2g_only: bool):
    available_p2g_names, available_custom_names, available_satpy_names = reader_info.get_available_products()
    available_satpy_names = ["*" + _sname for _sname in available_satpy_names]
    available_custom_names = ["*" + _sname for _sname in available_custom_names]
    project_name = "Polar2Grid" if is_polar2grid else "Geo2Grid"

    print("### Custom User Products")
    print("\n".join(sorted(available_custom_names)) if available_custom_names else "<None>")
    print()

    if not p2g_only:
        print("### Non-standard Satpy Products")
        print("\n".join(sorted(available_satpy_names)) if available_satpy_names else "<None>")
        print()

    print(f"### Standard Available {project_name} Products")
    print("\n".join(sorted(available_p2g_names)) if available_p2g_names else "<None>")


def _create_scene(scene_creation: dict) -> Optional[Scene]:
    try:
        scn = Scene(**scene_creation)
    except ValueError as e:
        LOG.error("{} | Enable debug message (-vvv) or see log file for details.".format(str(e)))
        LOG.debug("Further error information: ", exc_info=True)
        return
    except OSError:
        LOG.error("Could not open files. Enable debug message (-vvv) or see log file for details.")
        LOG.debug("Further error information: ", exc_info=True)
        return
    return scn


def _resample_scene_to_grids(
    scn: Scene,
    reader_names: list[str],
    resample_args: dict,
    filter_kwargs: dict,
    preserve_resolution: bool,
    use_polar2grid_defaults: bool,
) -> list[tuple]:
    ll_bbox = resample_args.pop("ll_bbox")
    if ll_bbox:
        scn = scn.crop(ll_bbox=ll_bbox)

    scn = filter_scene(
        scn,
        reader_names,
        **filter_kwargs,
    )
    if scn is None:
        LOG.info("No remaining products after filtering.")
        return []

    areas_to_resample = resample_args.pop("grids")
    antimeridian_mode = resample_args.pop("antimeridian_mode")
    if "ewa_persist" in resample_args:
        resample_args["persist"] = resample_args.pop("ewa_persist")
    scenes_to_save = resample_scene(
        scn,
        areas_to_resample,
        antimeridian_mode=antimeridian_mode,
        preserve_resolution=preserve_resolution,
        is_polar2grid=use_polar2grid_defaults,
        **resample_args,
    )
    return scenes_to_save


def _save_scenes(scenes_to_save: list[tuple], reader_info, writer_args) -> list:
    to_save = []
    for scene_to_save, products_to_save in scenes_to_save:
        _overwrite_platform_name_with_aliases(scene_to_save)
        _overwrite_sensor_with_aliases(scene_to_save)
        reader_info.apply_p2g_name_to_scene(scene_to_save)
        to_save = _write_scene(
            scene_to_save,
            writer_args["writers"],
            writer_args,
            products_to_save,
            to_save=to_save,
        )
    return to_save


def _get_glue_name(args):
    reader_name = "NONE" if args.readers is None else args.readers[0]
    writer_names = "-".join(args.writers or [])
    return f"{reader_name}_{writer_names}"


@contextlib.contextmanager
def _create_profile_html_if(create_profile: Union[False, None, str], project_name: str, glue_name: str):
    from dask.diagnostics import CacheProfiler, Profiler, ResourceProfiler, visualize

    if create_profile is False:
        yield
        return
    if create_profile is None:
        profile_filename = "{project_name}_{glue_name}_{start_time:%Y%m%d_%H%M%S}.html"
    else:
        profile_filename = create_profile

    start_time = datetime.now()
    with CacheProfiler() as cprof, ResourceProfiler() as rprof, Profiler() as prof:
        yield
    end_time = datetime.now()

    profile_filename = profile_filename.format(
        project_name=project_name,
        glue_name=glue_name,
        start_time=start_time,
        end_time=end_time,
    )
    profile_filename = os.path.abspath(profile_filename)
    visualize([prof, rprof, cprof], file_path=profile_filename, show=False)
    print(f"Profile HTML: file://{profile_filename}")


def main(argv=sys.argv[1:]):
    ret = -1
    try:
        processor = _GlueProcessor(argv)
    except FileNotFoundError:
        return ret

    try:
        ret = processor()
    finally:
        processor.cleanup()
    return ret


class _GlueProcessor:
    """Helper class to make calling processing steps easier."""

    def __init__(self, argv):
        add_polar2grid_config_paths()
        self.is_polar2grid = get_p2g_defaults_env_var()
        self.arg_parser = GlueArgumentParser(argv, self.is_polar2grid)
        self.glue_name = _get_glue_name(self.arg_parser._args)
        self.rename_log = _prepare_initial_logging(self.arg_parser, self.glue_name)
        self.tmp_config_paths = []
        self._handle_extra_config_paths(self.arg_parser._args)
        self._clean = False

    def _handle_extra_config_paths(self, args):
        if not args.extra_config_path:
            return
        _check_valid_config_paths(args.extra_config_path)

        new_config_paths = []
        # Preserve user's specified order to handle inheritance/overrides
        for extra_config_path in args.extra_config_path:
            if os.path.isdir(extra_config_path):
                new_config_paths.append(extra_config_path)
                continue

            tmp_config_path = _create_tmp_enhancement_config_dir(extra_config_path)
            new_config_paths.append(tmp_config_path)
            self.tmp_config_paths.append(tmp_config_path)
        _add_extra_config_paths(new_config_paths)
        LOG.debug(f"Satpy config path is: {satpy.config.get('config_path')}")

    def cleanup(self):
        self._clean = True
        for tmp_config_path in self.tmp_config_paths:
            LOG.debug(f"Deleting temporary config directory: {tmp_config_path}")
            shutil.rmtree(tmp_config_path, ignore_errors=True)

    def __call__(self):
        # Set up dask and the number of workers
        common_args = self.arg_parser._args
        if common_args.num_workers:
            dask.config.set(num_workers=common_args.num_workers)

        with _create_profile_html_if(
            common_args.create_profile,
            "polar2grid" if self.is_polar2grid else "geo2grid",
            self.glue_name,
        ):
            return self._run_processing()

    def _run_processing(self):
        LOG.info("Sorting and reading input files...")
        arg_parser = self.arg_parser
        preferred_chunk_size = get_reader_attr(arg_parser._scene_creation["reader"], "PREFERRED_CHUNK_SIZE", 1024)
        _set_preferred_chunk_size(preferred_chunk_size)
        scn = _create_scene(arg_parser._scene_creation)
        if scn is None:
            return -1

        if self.rename_log:
            stime = getattr(scn, "start_time", scn.attrs.get("start_time"))
            rename_log_file(self.glue_name + stime.strftime("_%Y%m%d_%H%M%S.log"))

        # Load the actual data arrays and metadata (lazy loaded as dask arrays)
        LOG.info("Loading product metadata from files...")
        load_args = arg_parser._load_args.copy()
        user_products = load_args.pop("products")
        reader_info = ReaderProxyBase.from_reader_name(arg_parser._scene_creation["reader"], scn, user_products)
        if arg_parser._args.list_products or arg_parser._args.list_products_all:
            _print_list_products(reader_info, self.is_polar2grid, not arg_parser._args.list_products_all)
            return 0

        products = reader_info.get_satpy_products_to_load()
        persist_geolocation = not arg_parser._reader_args.pop("no_persist_geolocation", False)
        if not products:
            return -1
        scn.load(products, **load_args, generate=False)
        if persist_geolocation:
            scn = _persist_swath_definition_in_scene(scn)
        scn.generate_possible_composites(True)

        reader_args = arg_parser._reader_args
        filter_kwargs = {
            "sza_threshold": reader_args["sza_threshold"],
            "day_fraction": reader_args["filter_day_products"],
            "night_fraction": reader_args["filter_night_products"],
        }
        scenes_to_save = _resample_scene_to_grids(
            scn,
            arg_parser._reader_names,
            arg_parser._resample_args,
            filter_kwargs,
            arg_parser._args.preserve_resolution,
            self.is_polar2grid,
        )
        to_save = _save_scenes(scenes_to_save, reader_info, arg_parser._writer_args)

        if arg_parser._args.progress:
            pbar = ProgressBar()
            pbar.register()

        LOG.info("Computing products and saving data to writers...")
        if not to_save:
            LOG.warning(
                "No product files produced given available valid data and "
                "resampling settings. This can happen if the writer "
                "detects that no valid output will be written or the "
                "input data does not overlap with the target grid."
            )
        compute_writer_results(to_save)
        LOG.info("SUCCESS")
        return 0


def _prepare_initial_logging(arg_parser, glue_name: str) -> bool:
    global LOG
    LOG = logging.getLogger(glue_name)

    # Prepare logging
    args = arg_parser._args
    rename_log = False
    if args.log_fn is None:
        rename_log = True
        args.log_fn = glue_name + "_fail.log"
    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    setup_logging(console_level=levels[min(3, args.verbosity)], log_filename=args.log_fn)
    logging.getLogger("rasterio").setLevel(levels[min(1, args.verbosity)])
    logging.getLogger("fsspec").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("s3fs").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("aiobotocore").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("botocore").setLevel(levels[min(2, args.verbosity)])
    sys.excepthook = create_exc_handler(LOG.name)
    if levels[min(3, args.verbosity)] > logging.DEBUG:
        import warnings

        warnings.filterwarnings("ignore")
    LOG.debug("Starting script with arguments: %s", " ".join(arg_parser.argv))
    return rename_log


def _add_extra_config_paths(extra_paths: list[str]):
    config_path = satpy.config.get("config_path")
    LOG.info(f"Adding additional configuration paths: {extra_paths}")
    satpy.config.set(config_path=config_path + extra_paths)


def _check_valid_config_paths(extra_config_paths: Iterable):
    single_file_configs = [config_path for config_path in extra_config_paths if os.path.isfile(config_path)]
    config_paths = [config_path for config_path in extra_config_paths if os.path.isdir(config_path)]
    invalid_paths = set(extra_config_paths) - (set(single_file_configs) | set(config_paths))
    if invalid_paths:
        str_paths = "\n\t".join(sorted(invalid_paths))
        msg = f"Specified extra config paths don't exist:\n\t{str_paths}"
        LOG.error(msg)
        raise FileNotFoundError(msg)


def _create_tmp_enhancement_config_dir(enh_yaml_file: str) -> str:
    config_dir = tempfile.mkdtemp(prefix="p2g_tmp_config")
    enh_dir = os.path.join(config_dir, "enhancements")
    LOG.debug(f"Creating temporary config directory for enhancement file {enh_yaml_file!r}: {enh_dir}")
    os.makedirs(enh_dir)

    new_enh_file = os.path.join(enh_dir, "generic.yaml")
    shutil.copy(enh_yaml_file, new_enh_file)
    return config_dir


def _set_preferred_chunk_size(preferred_chunk_size: int) -> None:
    pcs_in_mb = (preferred_chunk_size * preferred_chunk_size) * 8 // (1024 * 1024)
    if "PYTROLL_CHUNK_SIZE" not in os.environ:
        LOG.debug(f"Setting preferred chunk size to {preferred_chunk_size} pixels or {pcs_in_mb:d}MiB")
        os.environ["PYTROLL_CHUNK_SIZE"] = f"{preferred_chunk_size:d}"
        dask.config.set({"array.chunk-size": f"{pcs_in_mb:d}MiB"})
    else:
        LOG.debug(f"Using environment variable chunk size: {os.environ['PYTROLL_CHUNK_SIZE']}")


def _persist_swath_definition_in_scene(scn: Scene) -> None:
    to_persist_swath_defs = _swaths_to_persist(scn)
    if not to_persist_swath_defs:
        return scn

    to_update_data_arrays, to_persist_lonlats = zip(*to_persist_swath_defs.values(), strict=True)
    LOG.info("Loading swath geolocation into memory...")
    persisted_lonlats = dask.persist(*to_persist_lonlats)
    persisted_swath_defs = [SwathDefinition(plons, plats) for plons, plats in persisted_lonlats]
    new_scn = scn.copy()
    for arrays_to_update, persisted_swath_def in zip(to_update_data_arrays, persisted_swath_defs, strict=True):
        for array_to_update in arrays_to_update:
            array_to_update.attrs["area"] = persisted_swath_def
            new_scn._datasets[array_to_update.attrs["_satpy_id"]] = array_to_update
    LOG.debug(f"{len(to_persist_swath_defs)} unique swath definitions persisted")
    return new_scn


def _swaths_to_persist(scn: Scene) -> dict:
    to_persist_swath_defs = {}
    for data_arr in scn.values():
        swath_def = data_arr.attrs.get("area")
        if not isinstance(swath_def, SwathDefinition):
            continue
        this_swath_data_array_copies, _ = to_persist_swath_defs.setdefault(
            swath_def, ([], (swath_def.lons, swath_def.lats))
        )
        this_swath_data_array_copies.append(data_arr.copy())
    return to_persist_swath_defs


if __name__ == "__main__":
    sys.exit(main())



"""Various behavior test steps."""

import glob
import os
import re
import shutil
import subprocess
import tempfile

from behave import given, then, when

DTYPE_REGEX = re.compile(r"--dtype (P?<dtype>\S+) ")


@given("input data from {source}")
def step_impl_input_data(context, source):
    new_source = ""

    for f in source.split(" "):
        f = os.path.join(context.datapath, f)
        new_source += f + " "

        if "*" in os.path.basename(f):
            assert glob.glob(f), "Input files {} do  not exist".format(f)
            assert os.path.exists(os.path.dirname(f)), "Input folder {} does not exist".format(os.path.dirname(f))
        else:
            assert os.path.exists(f), "Input {} does not exist".format(f)

    context.source = new_source.rstrip()


@given("an empty working directory")
def step_impl_empty_work_dir(context):
    prefix = os.path.join(context.base_temp_dir, "p2g_tests__")
    scen_name = context.scenario.name
    if "@" in scen_name:
        # "outline" test
        outline_info = scen_name[scen_name.find("@") + 1 :].strip()
        outline_info = outline_info.replace(" ", "_").replace(".", "-").lower()
        prefix += outline_info + "__"

    context.temp_dir = tempfile.mkdtemp(prefix=prefix)


@given("input data is copied to the working directory")
def step_impl_copy_input_data(context):
    new_source = ""
    for input_dir in context.source.split(" "):
        for input_filename in os.listdir(input_dir):
            new_path = os.path.join(context.temp_dir, os.path.basename(input_filename))
            new_source += new_path + " "
            shutil.copy(os.path.join(input_dir, input_filename), new_path)

    context.source = new_source.rstrip()


@when("{command} runs")
def step_impl_run_command(context, command):
    context.script = command.split()[0]
    context.command = "datapath={}; /usr/bin/time {} {}".format(
        context.datapath, os.path.join(context.p2g_path, os.path.expandvars(command)), context.source
    )

    # creating new data in temporary directory to compare
    orig_dir = os.getcwd()
    temp_dir = getattr(context, "temp_dir", orig_dir)
    try:
        os.chdir(temp_dir)
        exit_status = subprocess.call(context.command, shell=True)
        assert exit_status == 0, "{} ran unsuccessfully".format(context.command)
    finally:
        os.chdir(orig_dir)


@then("the same filenames as in {output} were produced")
def step_impl_match_output_filenames(context, output):
    for exp_filename in os.listdir(os.path.join(context.datapath, output)):
        assert os.path.isfile(os.path.join(context.temp_dir, exp_filename))


@then("the output matches with the files in {output}")
def step_impl_match_output_files(context, output):
    orig_dir = os.getcwd()
    dtype = _get_dtype_from_command(context.command)
    try:
        os.chdir(context.datapath)
        # NOTE: 81231 / 151404144 (0.054%) pixels are currently wrong in VIIRS_L1B.
        compare_command = [
            "python3",
            "-m",
            "polar2grid.compare",
            output,
            context.temp_dir,
            "-vv",
            "--dtype",
            dtype or "float32",
            "--margin-of-error",
            str(81231 / 1514041.44),
        ]
        if context.html_dst:
            html_path = _html_output_filename(context, output)
            compare_command.extend(["--html", html_path])
        exit_status = subprocess.call(" ".join(compare_command), shell=True)
        assert exit_status == 0, "Files did not match with the correct output"
    finally:
        os.chdir(orig_dir)


def _html_output_filename(context, output_dir: str) -> str:
    html_fn = os.path.join(context.html_dst, "{}.html".format(output_dir))
    return os.path.join(context.temp_dir, html_fn)


@when("{command} runs with --list-products")
def step_impl_run_list_products(context, command):
    context.script = command.split()[0]
    context.command = "datapath={}; /usr/bin/time {} {} --list-products".format(
        context.datapath, os.path.join(context.p2g_path, command), context.source
    )

    # creating new data in temporary directory to compare
    orig_dir = os.getcwd()
    try:
        context.temp_dir = tempfile.mkdtemp(prefix=os.path.join(context.base_temp_dir, "p2g_tests_"))
        os.chdir(context.temp_dir)
        output = subprocess.check_output(context.command, shell=True)
        context.output = output.decode("utf-8")
    finally:
        os.chdir(orig_dir)

    assert os.path.exists(context.temp_dir), "Temporary directory not created"
    assert os.listdir(context.temp_dir), "No files were created"


@then("the printed output includes the products in {output}")
def step_impl_compare_output(context, output):
    orig_dir = os.getcwd()
    names_to_check = output.split(",")
    output = context.output
    try:
        for product_name in names_to_check:
            num_products_in_output = output.count(f"\n{product_name}\n")
            assert num_products_in_output != 0, f"Missing {product_name} in command output"
            assert num_products_in_output == 1, f"Too many of {product_name} in command output"
    finally:
        os.chdir(orig_dir)


def _get_dtype_from_command(cmd_str):
    match_res = DTYPE_REGEX.match(cmd_str)
    if match_res is None:
        return None
    return match_res.match_dict()["dtype"]



# -*- coding: utf-8 -*-
#
# polar2grid documentation build configuration file, created by
# sphinx-quickstart on Thu Apr 19 23:17:38 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
"""Polar2Grid project's sphinx documentation configuration."""

import ftplib
import os
import sys
import urllib.request
import warnings
from datetime import datetime
from shutil import copyfileobj

import yaml
from pyproj import CRS

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))
print("Adding the following directories to PYTHONPATH:")
BASE_PATH = "../../../"
for dirname in [
    x for x in os.listdir(BASE_PATH) if os.path.isdir(os.path.join(BASE_PATH, x)) and x.startswith("polar2grid")
]:
    print("\t ", os.path.realpath(os.path.join(BASE_PATH, dirname)))
    sys.path.insert(0, os.path.abspath(os.path.join(BASE_PATH, dirname)))
sys.path.append(os.path.abspath(os.path.dirname(__file__)))

# Handle building documentation for polar2grid or geo2grid
is_geo2grid = "geo" in os.getenv("POLAR2GRID_DOC", "polar").lower()

# Hack to download example images instead of storing them in git
images = (
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/amsr2/images_basic/gcom-w1_amsr2_btemp_36.5h_20160719_190300_wgs84_fit.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/amsr2/images_nrl/gcom-w1_amsr2_btemp_89.0ah_20160719_190300_lcc_fit.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/amsr2/images_nrl/gcom-w1_amsr2_btemp_89.0ah_20160719_190300_lcc_fit.basic_overlay_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/amsr2/images_nrl/gcom-w1_amsr2_btemp_89.0ah_20160719_190300_lcc_fit.advanced_overlay.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/overlay/npp_viirs_true_color_20170305_193251_lcc_fit_overlay.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/viirs/images_basic/noaa20_viirs_true_color_20220919_175331_lcc_fit.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/viirs/overlay/noaa20_viirs_true_color_20220919_175331_lcc_fit_overlay.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/dnb/VIIRS_DNB_Enhancement_Comparison.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/images_basic/npp_viirs_true_color_20170319_183246_miami.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/images_basic/terra_modis_false_color_20170319_163000_miami.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/modis/images_basic/terra_modis_false_color_20170319_163000_miami_p2g_v3.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/awips/SSEC_AWIPS_aqua_modis_bt28_211e_20170308_181800.nc.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/awips/SSEC_AWIPS_aqua_modis_vis02_211e_20170308_181800.nc.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/amsr2/images_nrl/gcom-w1_amsr2_btemp_89.0ah_20160719_190300_lcc_fit_color.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/merge/VIIRS_False_Color_Side_by_Side_Example_P2G.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/overlay/VIIRS_true_color_in_google_earth.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/viirs/images_basic/noaa20_viirs_true_color_20220919_175331_lcc_fit_in_google_earth.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/merge/my_false_color.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/awips/modis_true_color_awips_20170308_1818.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_LCC.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_Mercator.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_Pacific.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_Polar.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_GOES_EAST.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_GOES_STORE.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_GOES_TEST.png",
    "https://www.ssec.wisc.edu/~davidh/polar2grid/scmi_grids/scmi_grid_GOES_WEST.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/viirs/dnb/HNCC_DNB_Band_Example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/awips/modis_true_color_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/modis/awips/modis_vis02_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/VIIRS_I02_ACSPO_SST_composite_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/noaa20_viirs_sst_20220810_184327_great_lakes.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/noaa20_viirs_sst_20220810_184327_great_lakes_wcolor.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/noaa20_viirs_sst_20220810_184327_great_lakes_rescaled_wcolor.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/noaa20_viirs_sst_20220810_184327_great_lakes_rescaled_wcolor_colortable_resize.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/acspo/noaa20_viirs_sst_20220810_184327_great_lakes_sst_final_resize.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/asci/VIIRS_AOD_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/asci/VIIRS_AOD_example_with_overlays.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/asci/VIIRS_CTT_example_with_overlays.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/asci/VIIRS_NDVI_example_with_overlays.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/asci/VIIRS_SurfReflectance_True_Color_example_with_overlays.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/acspo/npp_viirs_sst_20191216_072134_acspo_sst_rescaled_wcolor.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/acspo/npp_viirs_sst_20191216_072134_acspo_sst_rescaled_wcolor_colortable_resize.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/acspo/npp_viirs_sst_20191216_072134_acspo_sst_final_resize.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/amsr2/gcom-w1_amsr2_btemp_89.0ah_20220910_233500_wgs84_fit.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_3_0_examples/amsr2/gcom-w1_amsr2_btemp_89.0ah_20220910_233500_wgs84_lcc_fit_rescaled_and_colorized.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/ahi/HIMAWARI-8_AHI_true_color_20181112_233020_perth_example.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/ahi/HIMAWARI-8_AHI_true_color_20181112_233020_perth.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/abi_20181219_1745_montage.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/abi_20181219_1745_montage.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadC_natural_color_20181219_174215_GOES-East.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-18_ABI_RadF_natural_color_20221115_183020_GOES-West_cutout.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadC_true_color_20190104_195718_GOES-East.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadF_true_color_night_20181112_123034_GOES-East.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadF_true_color_night_20181112_123034_GOES-East_new.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadF_true_color_night_microphysics_20181112_123034_GOES-East_newenhancement.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadF_true_color_night_microphysics_20181112_123034_GOES-East.jpg",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/GOES-16_ABI_RadF_true_color_night_microphysics_20181112_123034_GOES-East_new.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/my_goes16_abi_naturalcolor.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi/my_goes18_abi_naturalcolor.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/GOES-17_ABI_TEMP_20221123_183117_GOES-West_original.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/goes17_overlay_true_color_cloud_temperature.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/my_goes17_abi_ctt.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/my_goes17_abi_ctt1.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/GOES-16_ABI_AOD_20240429_225020_GOES-East.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/g2g_examples/abi_l2/GOES-16_ABI_AOD_20240429_225020_GOES-East_woverlays.png",
    "https://bin.ssec.wisc.edu/pub/CSPP/p2g_v_2_1_examples/flood/Flood_Legend.png",
)
script_path = os.path.dirname(os.path.realpath(__file__))
image_dst = os.path.join(script_path, "_static", "example_images")
os.makedirs(image_dst, exist_ok=True)

for image_url in images:
    image_fn = os.path.basename(image_url)
    image_pathname = os.path.join(image_dst, image_fn)
    if os.path.isfile(image_pathname):
        continue
    elif image_url.startswith("http://") or image_url.startswith("https://"):  # or image_url.startswith('ftp://'):
        print("Downloading example image: {}".format(image_url))
        with urllib.request.urlopen(image_url) as remote_img, open(image_pathname, "wb") as local_img:
            copyfileobj(remote_img, local_img)
    elif image_url.startswith("ftp://"):
        print("Downloading example image: {}".format(image_url))
        parts = image_url.split("/")
        server = parts[2]
        ftp_fn = "/".join(parts[3:])
        ftp = ftplib.FTP(server, user="ftp")  # hope for anonymous
        out_file = open(image_pathname, "wb")
        ftp.retrbinary("RETR {}".format(ftp_fn), out_file.write)
    else:
        raise ValueError("Not sure how to download image: {}".format(image_url))

# -- Customize setup -----------------------------------------------------------


def setup(app):
    app.add_css_file("prettytables.css")
    app.add_config_value("is_geo2grid", is_geo2grid, "env")


# -- General configuration -----------------------------------------------------
rst_epilog = """
.. |ssec| replace:: :abbr:`SSEC (Space Science and Engineering Center)`
.. |cspp| replace:: :abbr:`CSPP (Community Satellite Processing Package)`
.. |viirs| replace:: :abbr:`VIIRS (Visible/Infrared Imager Radiometer Suite)`
"""

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.imgmath",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinxarg.ext",
    "doi_role",
    "toctree_filter",
    "sphinxcontrib.apidoc",
]

# API docs
apidoc_module_dir = "../../polar2grid"
apidoc_output_dir = "dev_guide/api"
if is_geo2grid:
    apidoc_excluded_paths = ["readers/amsr2_l1b.py"]
else:
    apidoc_excluded_paths = []
apidoc_separate_modules = True
apidoc_extra_args = ["-P"]

# Autodoc
autodoc_mock_imports = []

numfig = True

# graphviz_dot = 'dot'

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
if is_geo2grid:
    project = "Geo2Grid"
    version = "1.2"
    release = "1.2"
else:
    project = "Polar2Grid"

    # The version info for the project you're documenting, acts as replacement for
    # |version| and |release|, also used in various other places throughout the
    # built documents.
    #
    # The short X.Y version.
    version = "3.1"

    # The full version, including alpha/beta/rc tags.
    release = "3.1"

# Tell Polar2Grid scripts which system we are running
os.environ["USE_POLAR2GRID_DEFAULTS"] = str(int(not is_geo2grid))

rst_epilog += """
.. |project| replace:: {}
.. |script| replace:: {}.sh
.. |script_literal| replace:: ``{}.sh``
.. |project_env| replace:: ${}_HOME
.. |cspp_abbr| replace:: {}
.. |cspp_title| replace:: {}
""".format(
    project,
    project.lower(),
    project.lower(),
    project.upper(),
    "CSPP Geo" if is_geo2grid else "CSPP LEO",
    "CSPP for Geostationary Satellites" if is_geo2grid else "CSPP for Low Earth Orbiter Satellites",
)

# Custom TOCTree filtering
toc_filter_exclude = ["polar2grid"] if is_geo2grid else ["geo2grid"]

copyright = "2012-{:%Y}, University of Wisconsin SSEC".format(datetime.utcnow())

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    "readers/viirs_edr_flood.rst",  # not advertised
    "grids_list.rst",  # included directly in grids.rst
]
if is_geo2grid:
    exclude_patterns.extend(
        [
            "examples/acspo_example.rst",
            "examples/amsr2_example.rst",
            "examples/modis_example.rst",
            "examples/viirs_example.rst",
            "examples/asci_example.rst",
            "readers/acspo.rst",
            "readers/amsr2_l1b.rst",
            "readers/avhrr.rst",
            "readers/clavrx.rst",
            "readers/mersi2_l1b.rst",
            "readers/mersi_ll_l1b.rst",
            "readers/mirs.rst",
            "readers/modis_l1b.rst",
            "readers/modis_l2.rst",
            "readers/nucaps.rst",
            "readers/viirs_edr_active_fires.rst",
            "readers/viirs_l1b.rst",
            "readers/viirs_sdr.rst",
            "readers/viirs_edr.rst",
            "readers/virr_l1b.rst",
            "verification/modis_verification.rst",
            "verification/viirs_verification.rst",
            "viirs_day_night_band.rst",
            "writers/awips_tiled.rst",
            "writers/binary.rst",
            "writers/hdf5.rst",
            "misc_recipes.rst",
        ]
    )
else:
    exclude_patterns.extend(
        [
            "compositors.rst",
            "data_access.rst",
            "examples/abi_example.rst",
            "examples/abi_l2_example.rst",
            "examples/ahi_example.rst",
            "examples/creating_animations_example.rst",
            "readers/abi_l1b.rst",
            "readers/abi_l2_nc.rst",
            "readers/agri_fy4a_l1.rst",
            "readers/agri_fy4b_l1.rst",
            "readers/ahi_hrit.rst",
            "readers/ahi_hsd.rst",
            "readers/ami_l1b.rst",
            "readers/glm_l2.rst",
            "readers/fci_l1c_nc.rst",
            "verification/abi_verification.rst",
        ]
    )

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/SSEC_logo_small24.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = False

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = project.lower() + "doc"


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    "preamble": r"""
\setcounter{tocdepth}{1}
\usepackage{pdflscape}

\def\ttl@save@mkschap #1{\vspace *{-20\p@ }{\parindent \z@ \raggedright
    \normalfont \interlinepenalty \@M \DOTIS {#1} \vskip -20\p@ }}
\makeatother

""",
    "classoptions": ",openany,oneside",
    "babel": "\\usepackage[english]{babel}",
    "printindex": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]), toctree_only.
if is_geo2grid:
    latex_documents = [
        (
            "index",
            "{}_Documentation_{}.tex".format(project, version),
            "{} Documentation".format(project),
            "NOAA Community Satellite Processing Package \\and for Geostationary Satellites (CSPP Geo)",
            "manual",
            True,
        ),
    ]
else:
    latex_documents = [
        (
            "index",
            "{}_Documentation_{}.tex".format(project, version),
            "{} Documentation".format(project),
            "Released through the \\and NOAA Community Satellite Processing Package (CSPP)",
            "manual",
            True,
        ),
    ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "_static/{}2G_PDF_Logos.png".format("G" if is_geo2grid else "P")

# latex_toplevel_sectioning = 'section'

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
if is_geo2grid:
    latex_appendices = [
        # "misc_recipes",
        "design_overview",
    ]
else:
    latex_appendices = [
        "version3_implementation",
        "misc_recipes",
        "design_overview",
        "viirs_day_night_band",
    ]

# If false, no module index is generated.
latex_domain_indices = False

latex_engine = "xelatex"


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", project.lower(), "{} Documentation".format(project), ["David Hoese"], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        project.lower(),
        "{} Documentation".format(project),
        "David Hoese",
        project.lower(),
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "pyresample": ("https://pyresample.readthedocs.io/en/stable", None),
    "trollsift": ("https://trollsift.readthedocs.io/en/stable", None),
    "satpy": ("https://satpy.readthedocs.io/en/stable", None),
    "trollimage": ("https://trollimage.readthedocs.io/en/stable", None),
}

# Generate builtin grids list

grid_titles = {
    "wgs84_fit": "WGS84 Dynamic Fit",
    "wgs84_fit_250": "WGS84 Dynamic Fit 250m",
    "lcc_fit": "Lambert Conic Conformal Dynamic Fit",
    "lcc_fit_hr": "High Resolution Lambert Conic Conformal Dynamic Fit",
    "lcc_sa": "Lambert Conic Conformal - South America Centered",
    "lcc_eu": "Lambert Conic Conformal - Europe Centered",
    "lcc_south_africa": "Lambert Conic Conformal - South Africa Centered",
    "lcc_aus": "Lambert Conic Conformal - Australia Centered",
    "lcc_asia": "Lambert Conic Conformal - Asia Centered",
    "polar_north_pacific": "Polar-Stereographic North Pacific",
    "polar_south_pacific": "Polar-Stereographic South Pacific",
    "polar_alaska": "Polar-Stereographic Alaska",
    "polar_canada": "Polar-Stereographic Canada",
    "polar_russia": "Polar-Stereographic Russia",
    "eqc_fit": "Equirectangular Fit",
    "goes_east_1km": "GOES-East 1km",
    "goes_east_4km": "GOES-East 4km",
    "goes_east_8km": "GOES-East 8km",
    "goes_east_10km": "GOES-East 10km",
    "goes_west_1km": "GOES-West 1km",
    "goes_west_4km": "GOES-West 4km",
    "goes_west_8km": "GOES-West 8km",
    "goes_west_10km": "GOES-West 10km",
}

warnings.filterwarnings("ignore", module="pyproj", category=UserWarning)
builtin_areas_filename = os.path.join(script_path, "..", "..", "polar2grid", "grids", "grids.yaml")
with open(builtin_areas_filename, "r") as yaml_file:
    areas_dict = yaml.load(yaml_file, Loader=yaml.SafeLoader)
# builtin_areas = parse_area_file(os.path.join(script_path, "..", "..", "polar2grid", "grids", "grids.yaml"))
grids_list_filename = os.path.join(script_path, "grids_list.rst")
with open(grids_list_filename, "w") as grids_list_file:
    for area_name, area_dict in areas_dict.items():
        area_title = grid_titles.get(area_name)
        if area_title is None:
            continue
        proj = area_dict["projection"]
        crs = CRS.from_user_input(proj["EPSG"] if "EPSG" in proj else proj)
        title_underline = "^" * len(area_title)
        rst_str = f"""
.. _grid_{area_name}:

{area_title}
{title_underline}

:Grid Name: {area_name}
:Description: {area_dict['description']}
:Projection: {crs.to_string()}
"""

        if "resolution" in area_dict:
            res = area_dict["resolution"]
            xres = res["dx"]
            yres = res["dy"]
            def_units = "degrees" if crs.is_geographic else "meters"
            units = res.get("units", def_units)
            if xres != yres:
                rst_str += f":Resolution (X): {xres} {units}\n"
                rst_str += f":Resolution (Y): {yres} {units}\n"
            else:
                rst_str += f":Resolution: {xres} {units}\n"
        if "area_extent" in area_dict:
            rst_str += f":Extent: {area_dict['area_extent']}\n"
        grids_list_file.write(rst_str)



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2016-2022 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Add a colortable to an existing GeoTIFF."""

import os
import sys

import numpy as np
import rasterio
from trollimage.colormap import Colormap


def load_color_table_file_to_colormap(ct_file):
    """Load a text colormap file and create a trollimage.Colormap object.

    If the provided pathname includes ``$POLAR2GRID_HOME`` it will be replaced
    with the current environment variable value of ``POLAR2GRID_HOME``.

    """
    from satpy.enhancements import create_colormap

    p2g_home = os.getenv("POLAR2GRID_HOME", "")
    ct_file = ct_file.replace("$POLAR2GRID_HOME", p2g_home)
    ct_file = ct_file.replace("$GEO2GRID_HOME", p2g_home)
    cmap = create_colormap({"filename": ct_file})
    return cmap


def create_colortable(ct_file_or_entries):
    """Create GDAL ColorTable object from Colormap object."""
    if isinstance(ct_file_or_entries, str):
        ct_file_or_entries = load_color_table_file_to_colormap(ct_file_or_entries)
    if isinstance(ct_file_or_entries, Colormap):
        values = np.arange(256)
        new_colors = ct_file_or_entries.colorize(values)
        new_colors = np.clip(new_colors, 0.0, 1.0)
        ct_file_or_entries = (
            (color_idx,) + tuple(np.round(new_colors[:, color_idx] * 255.0)) for color_idx in range(new_colors.shape[1])
        )

    ct_dict = {entry[0]: tuple(entry[1:]) for entry in ct_file_or_entries}
    return ct_dict


def add_colortable(gtiff_ds, ct_dict):
    num_bands = gtiff_ds.count
    if num_bands in (2, 4):
        # don't add a color table to alpha bands
        num_bands -= 1
    for band_num in range(num_bands):
        gtiff_ds.write_colormap(band_num + 1, ct_dict)


def get_parser():
    import argparse

    description = "Add a GeoTIFF colortable to an existing single-band GeoTIFF."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("ct_file", help="Color table file to apply (CSV of (int, R, G, B, A)")
    parser.add_argument("geotiffs", nargs="+", help="Geotiff files to apply the color table to")
    return parser


def main(argv=sys.argv[1:]):
    parser = get_parser()
    args = parser.parse_args(argv)

    ct = create_colortable(args.ct_file)
    for geotiff_fn in args.geotiffs:
        gtiff_ds = rasterio.open(geotiff_fn, "r+")
        add_colortable(gtiff_ds, ct)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2012-2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The MiRS frontend extracts data from files created by the Community
Satellite Processing Package (CSPP) direct broadcast version of the
NOAA/STAR Microwave integrated Retrieval System (MiRS). The software
supports the creation of atmospheric and surface parameters from
ATMS, AMSU-A, and MHS microwave sensor data. For more information
on this product, please visit the CSPP LEO website:
`https://cimss.ssec.wisc.edu/cspp/`.

When executed on  Advanced Technology Microwave Sounder (ATMS)
MiRS product files, a limb correction algorithm is applied for
brightness temperatures reprojections for each of the 22 spectral bands.
The correction software was provided by Kexin Zhang of NOAA STAR, and
is applied as part of the MIRS ATMS Polar2Grid execution.

This reader's default resampling algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--weight-delta-max`` option is set to 100 and the
``--weight-distance-max`` option is set to 1.

The ACSPO output product format is NetCDF4. The frontend can be
specified with the `polar2grid.sh` command using the `mirs`
frontend name.The frontend offers the following products:

    +--------------------+----------------------------------------------------+
    | Product Name       | Description                                        |
    +====================+====================================================+
    | rain_rate          | Rain Rate                                          |
    +--------------------+----------------------------------------------------+
    | sea_ice            | Sea Ice in percent                                 |
    +--------------------+----------------------------------------------------+
    | snow_cover         | Snow Cover                                         |
    +--------------------+----------------------------------------------------+
    | tpw                | Total Precipitable Water                           |
    +--------------------+----------------------------------------------------+
    | swe                | Snow Water Equivalence                             |
    +--------------------+----------------------------------------------------+
    | clw                | Cloud Liquid Water                                 |
    +--------------------+----------------------------------------------------+
    | sfr                | Snow Fall Rate                                     |
    +--------------------+----------------------------------------------------+
    | surface_type       | Surface Type                                       |
    +--------------------+----------------------------------------------------+
    | tskin              | Skin Temperature                                   |
    +--------------------+----------------------------------------------------+
    | btemp_X            | Brightness Temperature for channel X (see below)   |
    +--------------------+----------------------------------------------------+

For specific brightness temperature band products, use the ``btemp_X`` option,
where ``X`` is a combination of the microwave frequency (integer) and
polarization of the channel. If there is more than one channel at that
frequency and polarization a "count" number is added to the end. To create
output files of all available bands, use the ``--bt-channels`` option.

As an example, the ATMS band options are:

.. list-table:: ATMS Brightness Temperature Channels
    :header-rows: 1

    * - **Band Number**
      - **Frequency (GHz)**
      - **Polarization**
      - **Polar2Grid Dataset Name**
    * - 1
      - 23.79
      - V
      - btemp_23v
    * - 2
      - 31.40
      - V
      - btemp_31v
    * - 3
      - 50.30
      - H
      - btemp_50h
    * - 4
      - 51.76
      - H
      - btemp_51h
    * - 5
      - 52.80
      - H
      - btemp_52h
    * - 6
      - 53.59±0.115
      - H
      - btemp_53h
    * - 7
      - 54.40
      - H
      - btemp_54h1
    * - 8
      - 54.94
      - H
      - btemp_54h2
    * - 9
      - 55.50
      - H
      - btemp_55h
    * - 10
      - 57.29
      - H
      - btemp_57h1
    * - 11
      - 57.29±2.17
      - H
      - btemp_57h2
    * - 12
      - 57.29±0.3222±0.048
      - H
      - btemp_57h3
    * - 13
      - 57.29±0.3222±0.022
      - H
      - btemp_57h4
    * - 14
      - 57.29±0.3222±0.010
      - H
      - btemp_57h5
    * - 15
      - 57.29±0.3222±0.0045
      - H
      - btemp_57h6
    * - 16
      - 88.20
      - V
      - btemp_88v
    * - 17
      - 165.50
      - H
      - btemp_165h
    * - 18
      - 183.31±7.0
      - H
      - btemp_183h1
    * - 19
      - 183.31±4.5
      - H
      - btemp_183h2
    * - 20
      - 183.31±3.0
      - H
      - btemp_183h3
    * - 21
      - 183.31±1.8
      - H
      - btemp_183h4
    * - 22
      - 183.31±1.0
      - H
      - btemp_183h5

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup, BooleanOptionalAction
from typing import Optional, Union

from satpy import DataID, Scene

from polar2grid.core.script_utils import ExtendConstAction

from ._base import ReaderProxyBase

PRECIP_PRODUCTS = [
    "rain_rate",
    "tpw",
    "clw",
]
SNOW_PRODUCTS = [
    "snow_cover",
    "swe",
    "sfr",
]
SEAICE_PRODUCTS = [
    "sea_ice",
]

PRODUCT_ALIASES = {}
PRODUCT_ALIASES["rain_rate"] = "RR"
PRODUCT_ALIASES["tpw"] = "TPW"
PRODUCT_ALIASES["clw"] = "CLW"

PRODUCT_ALIASES["snow_cover"] = "Snow"
PRODUCT_ALIASES["swe"] = "SWE"
PRODUCT_ALIASES["sfr"] = "SFR"

PRODUCT_ALIASES["sea_ice"] = "SIce"

PRODUCT_ALIASES["surface_type"] = "Sfc_type"
PRODUCT_ALIASES["tskin"] = "TSkin"

P2G_PRODUCTS = PRECIP_PRODUCTS + SNOW_PRODUCTS + SEAICE_PRODUCTS + ["surface_type", "tskin"]


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def __init__(self, scn: Scene, user_products: list[str]):
        self._modified_aliases = PRODUCT_ALIASES.copy()
        if "all_bt_channels" in user_products:
            # they specified --bt-channels
            user_products.remove("all_bt_channels")
            bt_channels = self._btemp_channels_from_satpy(scn.all_dataset_names())
            user_products.extend(set(bt_channels) - set(user_products))
        super().__init__(scn, user_products)

    @staticmethod
    def _btemp_channels_from_satpy(satpy_ids: list[Union[str, DataID]]):
        for data_id in satpy_ids:
            dname = data_id if isinstance(data_id, str) else data_id["name"]
            if dname.startswith("btemp_"):
                yield dname

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        all_btemps = list(self._btemp_channels_from_satpy(self.scn.all_dataset_names()))
        defaults = ["rain_rate"]
        defaults += set(all_btemps) & set(["btemp_88v", "btemp_89v1"])
        return defaults

    def get_all_products(self):
        """Get all polar2grid products that could be loaded."""
        all_btemps = list(self._btemp_channels_from_satpy(self.scn.all_dataset_names()))
        return P2G_PRODUCTS + all_btemps

    @property
    def _aliases(self):
        return PRODUCT_ALIASES


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser."""
    if group is None:
        group = parser.add_argument_group(title="MiRS Reader")

    group.add_argument(
        "--bt-channels",
        dest="products",
        action=ExtendConstAction,
        const=["all_bt_channels"],
        help="Add all bands to list of products",
    )
    group.add_argument(
        "--limb-correction",
        default=True,
        action=BooleanOptionalAction,
        help="Perform limb correction on ATMS instrument BT channels (enabled by default)",
    )
    return group, None
